# Copyright 2024 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Any, Callable

from secretflow.device.driver import PYU, SPU, DeviceObject, PYUObject, SPUObject, wait
from secretflow.spec.extend.data_pb2 import DeviceObjectCollection
from secretflow.spec.v1.data_pb2 import DistData, SystemInfo
from secretflow.utils import secure_pickle as pickle

from ..storage import Storage
from .base import IDumper


class Model(IDumper):
    def __init__(
        self,
        name: str,
        type: str,
        system_info: SystemInfo = None,  # type: ignore
        major_version: int = 0,
        minor_version: int = 0,
        public_info: Any = None,
        objs: list[DeviceObject] = None,
    ):
        self.name = name
        self.type = type
        self.system_info = system_info
        self.major_version = major_version
        self.minor_version = minor_version
        self.public_info = public_info
        self.objs = objs

    def check(
        self,
        model_type: str = None,
        max_major_version: int = None,
        max_minor_version: int = None,
    ):
        assert (
            model_type and model_type == self.type
        ), f"model type mismatch, {model_type}, {self.type}"
        assert (
            max_major_version and max_major_version >= self.major_version
        ), f"major version mismatch, {max_major_version}, {self.major_version}"
        assert (
            max_major_version and max_minor_version >= self.minor_version
        ), f"minor version mismatch, {max_minor_version}, {self.minor_version}"

    @staticmethod
    def load(storage: Storage, dist_data: DistData, spu: SPU | Callable[[], SPU] = None) -> 'Model':  # type: ignore
        model_meta = DeviceObjectCollection()
        assert dist_data.meta.Unpack(model_meta)

        model_info = json.loads(model_meta.public_info)

        assert (
            isinstance(model_info, dict)
            and "major_version" in model_info
            and "minor_version" in model_info
            and "public_info" in model_info
        )

        objs = []
        for save_obj in model_meta.objs:
            if save_obj.type == "pyu":
                assert len(save_obj.data_ref_idxs) == 1
                data_ref = dist_data.data_refs[save_obj.data_ref_idxs[0]]
                party = data_ref.party
                pyu = PYU(party)

                assert data_ref.format == "pickle"

                def loads(storage: Storage, path: str) -> Any:
                    with storage.get_reader(path) as r:
                        return pickle.load(r)

                objs.append(pyu(loads)(storage, data_ref.uri))
            elif save_obj.type == "spu":
                # TODO: only support one spu for now
                assert spu is not None
                if callable(spu):
                    spu = spu()
                assert len(save_obj.data_ref_idxs) > 1
                full_paths = {}
                for data_ref_idx in save_obj.data_ref_idxs:
                    data_ref = dist_data.data_refs[data_ref_idx]
                    assert data_ref.format == "pickle"
                    party = data_ref.party
                    assert party not in full_paths
                    uri = data_ref.uri
                    full_paths[party] = lambda uri=uri: storage.get_reader(uri)
                assert set(full_paths.keys()) == set(spu.actors.keys())
                spu_paths = [full_paths[party] for party in spu.actors.keys()]
                objs.append(spu.load(spu_paths))
            else:
                raise RuntimeError(f"not supported objs type {save_obj.type}")

        return Model(
            name=dist_data.name,
            type=dist_data.type,
            system_info=dist_data.system_info,
            public_info=dist_data.public_info,
            major_version=model_info["major_version"],
            minor_version=model_info["minor_version"],
            objs=objs,
        )

    def dump(self, storage: Storage, output_uris: str) -> DistData:  # type: ignore
        assert output_uris != "", f"output_uris cannot be empty"
        assert self.name != "", f"name is not set, uri is {output_uris}"
        assert self.type != "", f"type is not set, uri is {output_uris}"

        objs_uri = []
        objs_party = []
        saved_objs = []
        for i, obj in enumerate(self.objs):
            if isinstance(obj, PYUObject):
                device: PYU = obj.device
                uri = f"{output_uris}/{i}"

                def dumps(comp_storage, uri: str, obj: Any):
                    with comp_storage.get_writer(uri) as w:
                        pickle.dump(obj, w)

                wait(device(dumps)(storage, uri, obj))

                saved_obj = DeviceObjectCollection.DeviceObject(
                    type="pyu", data_ref_idxs=[len(objs_uri)]
                )
                saved_objs.append(saved_obj)
                objs_uri.append(uri)
                objs_party.append(device.party)
            elif isinstance(obj, SPUObject):
                device: SPU = obj.device
                uris = [f"{output_uris}/{i}" for _ in device.actors]

                device.dump(
                    obj,
                    [lambda uri=uri: storage.get_writer(uri) for uri in uris],
                )

                saved_obj = DeviceObjectCollection.DeviceObject(
                    type="spu",
                    data_ref_idxs=[len(objs_uri) + p for p in range(len(uris))],
                )
                saved_objs.append(saved_obj)
                objs_uri.extend(uris)
                objs_party.extend(list(device.actors.keys()))
            else:
                raise RuntimeError(f"not supported objs type {type(obj)}")

        model_info = {
            "major_version": self.major_version,
            "minor_version": self.minor_version,
            "public_info": self.public_info,
        }

        meta = DeviceObjectCollection(
            objs=saved_objs,
            public_info=json.dumps(model_info),
        )

        dd = DistData(
            name=self.name,
            type=str(self.type),
            system_info=self.system_info,
            data_refs=[
                DistData.DataRef(uri=uri, party=p, format="pickle")
                for uri, p in zip(objs_uri, objs_party)
            ],
        )
        dd.meta.Pack(meta)
        return dd
